@@ -7,8 +7,11 @@ authors under the ISC license.
 
 from __future__ import annotations
 
+import hashlib
 import itertools
 import logging
+from abc import ABC
+from dataclasses import dataclass
 from email.message import EmailMessage, Message
 from email.parser import BytesParser
 from io import BytesIO
@@ -30,7 +33,8 @@ from packaging.version import Version
 from resolvelib.providers import AbstractProvider
 from resolvelib.resolvers import RequirementInformation
 
-from pip_audit._cache import caching_session
+from pip_audit._dependency_source import RequirementHashes, UnsupportedHashAlgorithm
+from pip_audit._service import SkippedDependency
 from pip_audit._state import AuditState
 from pip_audit._util import python_version
 from pip_audit._virtual_env import VirtualEnv, VirtualEnvError
@@ -41,23 +45,35 @@ logger = logging.getLogger(__name__)
 PYTHON_VERSION: Version = python_version()
 
 
-class Candidate:
+@dataclass
+class Candidate(ABC):
     """
-    Represents a dependency candidate. A dependency being resolved may have
-    multiple candidates, which go through a selection process guided by various
+    Represents a dependency candidate.
+    """
+
+    name: str
+
+
+class ResolvedCandidate(Candidate):
+    """
+    Represents a resolved dependency candidate that has been retrieved from PyPI. A dependency being
+    resolved may have multiple candidates, which go through a selection process guided by various
     weights (version, `sdist` vs. `wheel`, etc.)
     """
 
     def __init__(
         self,
         name: str,
+        project: str,
         filename: Path,
         version: Version,
         *,
         url: str,
         extras: set[str],
         is_wheel: bool,
+        reqs: list[Requirement],
         session: CacheControl,
+        req_hashes: RequirementHashes,
         timeout: int | None = None,
         state: AuditState = AuditState(),
     ) -> None:
@@ -66,17 +82,21 @@ class Candidate:
         """
 
         self.name = canonicalize_name(name)
+        self.project = project
         self.filename = filename
         self.version = version
         self.url = url
         self.extras = extras
         self.is_wheel = is_wheel
+        self.reqs = reqs
         self._session = session
+        self.req_hashes = req_hashes
         self._timeout = timeout
         self._state = state
 
         self._metadata: Message | None = None
         self._dependencies: list[Requirement] | None = None
+        self.dist_hashes: dict[str, str] = {}
 
     def __repr__(self) -> str:  # pragma: no cover
         """
@@ -95,10 +115,15 @@ class Candidate:
         if self._metadata is None:
             self._state.update_state(f"Fetching metadata for {self.name} ({self.version})")
 
+            response = self._session.get(self.url, timeout=self._timeout)
+            response.raise_for_status()
+            candidate_data = response.content
+            self._generate_dist_hashes(candidate_data)
+
             if self.is_wheel:
-                self._metadata = self._get_metadata_for_wheel()
+                self._metadata = self._get_metadata_for_wheel(candidate_data)
             else:
-                self._metadata = self._get_metadata_for_sdist()
+                self._metadata = self._get_metadata_for_sdist(candidate_data)
         return self._metadata
 
     def _get_dependencies(self) -> Iterator[Requirement]:
@@ -126,15 +151,13 @@ class Candidate:
             self._dependencies = list(self._get_dependencies())
         return self._dependencies
 
-    def _get_metadata_for_wheel(self) -> Message:
+    def _get_metadata_for_wheel(self, wheel_data: bytes) -> Message:
         """
         Extracts the metadata for this candidate, if it's a wheel.
         """
-        data = self._session.get(self.url, timeout=self._timeout).content
-
         self._state.update_state(f"Extracting wheel for {self.name} ({self.version})")
 
-        with ZipFile(BytesIO(data)) as z:
+        with ZipFile(BytesIO(wheel_data)) as z:
             for n in z.namelist():
                 if n.endswith(".dist-info/METADATA"):
                     p = BytesParser()
@@ -145,15 +168,13 @@ class Candidate:
         # If we didn't find the metadata, return an empty dict
         return EmailMessage()  # pragma: no cover
 
-    def _get_metadata_for_sdist(self) -> Message:
+    def _get_metadata_for_sdist(self, sdist_data: bytes) -> Message:
         """
         Extracts the metadata for this candidate, if it's a source distribution.
         """
 
-        response: requests.Response = self._session.get(self.url, timeout=self._timeout)
-        response.raise_for_status()
-        sdist_data = response.content
         metadata = EmailMessage()
+        self._generate_dist_hashes(sdist_data)
 
         with TemporaryDirectory() as pkg_dir:
             sdist = Path(pkg_dir) / self.filename.name
@@ -188,22 +209,59 @@ class Candidate:
 
         return metadata
 
+    def _generate_dist_hashes(self, dist_data: bytes) -> None:
+        """
+        Populate a mapping of hash algorithm names to hashes for the candidate.
+        """
+        if self.project not in self.req_hashes:
+            return
+        hash_algorithms = self.req_hashes.supported_algorithms(self.project)
+        dist_hashes = {}
+        for algorithm in hash_algorithms:
+            if algorithm not in hashlib.algorithms_available:
+                raise UnsupportedHashAlgorithm(
+                    f"encountered hash with unknown algorithm: {algorithm}"
+                )
+            if algorithm not in hashlib.algorithms_guaranteed:  # pragma: no cover
+                raise UnsupportedHashAlgorithm(
+                    f"encountered hash with known but non-guaranteed algorithm: {algorithm}, this "
+                    "won't necessarily work on other platforms"
+                )
+            hasher = hashlib.new(algorithm, dist_data)
+            dist_hashes[algorithm] = hasher.hexdigest()
+        self.dist_hashes = dist_hashes
+
+
+@dataclass
+class SkippedCandidate(Candidate):
+    """
+    Represents a skipped dependency candidate. When a project can't be found on PyPI, we propagate a
+    `SkippedCandidate` to satisfy the dependency resolver and to signal to the caller that the
+    candidate couldn't be resolved properly.
+    """
+
+    skip_reason: str
+
 
 def get_project_from_indexes(
     index_urls: list[str],
     session: CacheControl,
     project: str,
+    reqs: list[Requirement],
     extras: set[str],
+    req_hashes: RequirementHashes,
     timeout: int | None,
     state: AuditState,
-) -> Iterator[Candidate]:
+) -> Iterator[ResolvedCandidate]:
     """Return candidates from all indexes created from the project name and extras."""
     project_found = False
     for index_url in index_urls:
         # Not all indexes are guaranteed to have the project so this isn't an error
         # We should only return an error if it can't be found on ANY of the supplied index URLs
         try:
-            yield from get_project_from_index(index_url, session, project, extras, timeout, state)
+            yield from get_project_from_index(
+                index_url, session, project, reqs, extras, req_hashes, timeout, state
+            )
             project_found = True
         except PyPINotFoundError:
             pass
@@ -217,10 +275,12 @@ def get_project_from_index(
     index_url: str,
     session: CacheControl,
     project: str,
+    reqs: list[Requirement],
     extras: set[str],
+    req_hashes: RequirementHashes,
     timeout: int | None,
     state: AuditState,
-) -> Iterator[Candidate]:
+) -> Iterator[ResolvedCandidate]:
     """Return candidates from an index created from the project name and extras."""
 
     # NOTE: The trailing slash is important here: without it, the `urljoin`
@@ -279,16 +339,19 @@ def get_project_from_index(
                 is_wheel = False
 
             # TODO: Handle compatibility tags?
-            yield Candidate(
+            yield ResolvedCandidate(
                 name,
+                project,
                 Path(filename),
                 version,
                 url=dist_url,
                 extras=extras,
                 is_wheel=is_wheel,
+                reqs=reqs,
                 timeout=timeout,
                 state=state,
                 session=session,
+                req_hashes=req_hashes,
             )
         except Exception:
             continue
@@ -303,8 +366,9 @@ class PyPIProvider(AbstractProvider):
     def __init__(
         self,
         index_urls: list[str],
+        req_hashes: RequirementHashes,
+        session: requests.Session,
         timeout: int | None = None,
-        cache_dir: Path | None = None,
         state: AuditState = AuditState(),
     ):
         """
@@ -312,6 +376,9 @@ class PyPIProvider(AbstractProvider):
 
         `index_urls` is a list of package index URLs.
 
+        `req_hashes` is a `RequirementHashes` to control what package/algorithm combinations we
+        generate hashes for.
+
         `timeout` is an optional argument to control how many seconds the component should wait for
         responses to network requests.
 
@@ -324,9 +391,11 @@ class PyPIProvider(AbstractProvider):
         index_urls = [url if url.endswith("/") else f"{url}/" for url in index_urls]
 
         self.index_urls = index_urls
+        self.req_hashes = req_hashes
+        self.session = session
         self.timeout = timeout
-        self.session = caching_session(cache_dir, use_pip=True)
         self._state = state
+        self.skip_deps: list[SkippedDependency] = []
 
     def identify(self, requirement_or_candidate: Requirement | Candidate) -> str:
         """
@@ -360,6 +429,9 @@ class PyPIProvider(AbstractProvider):
         self._state.update_state(f"Resolving {identifier}")
 
         requirements = list(requirements[identifier])
+        logger.debug(
+            f"{identifier} req specifier constraints: {[r.specifier for r in requirements]}"
+        )
 
         bad_versions = {c.version for c in incompatibilities[identifier]}
 
@@ -371,26 +443,48 @@ class PyPIProvider(AbstractProvider):
         # Need to pass the extras to the search, so they
         # are added to the candidate at creation - we
         # treat candidates as immutable once created.
-        candidates = sorted(
-            [
-                candidate
-                for candidate in get_project_from_indexes(
-                    self.index_urls, self.session, identifier, extras, self.timeout, self._state
-                )
-                if candidate.version not in bad_versions
-                and all(candidate.version in r.specifier for r in requirements)
-                # HACK(ww): Additionally check that each candidate's name matches the
-                # expected project name (identifier).
-                # This technically shouldn't be required, but parsing distribution names
-                # from package indices is imprecise/unreliable when distribution filenames
-                # are PEP 440 compliant but not normalized.
-                # See: https://github.com/pypa/packaging/issues/527
-                and candidate.name == identifier
-            ],
-            key=attrgetter("version", "is_wheel"),
-            reverse=True,
+        all_candidates = get_project_from_indexes(
+            self.index_urls,
+            self.session,
+            identifier,
+            requirements,
+            extras,
+            self.req_hashes,
+            self.timeout,
+            self._state,
         )
 
+        candidates: list[ResolvedCandidate]
+        try:
+            candidates = sorted(
+                [
+                    candidate
+                    for candidate in all_candidates
+                    if candidate.version not in bad_versions
+                    # NOTE(ww): We use `filter(...)` instead of checking
+                    # `candidate.version in r.specifier` because the former has subtle (and PEP 440
+                    # mandated) behavior around prereleases. Specifically, `filter(...)`
+                    # returns prereleases even if not explicitly configured, but only if
+                    # there are no non-prereleases.
+                    # See: https://github.com/pypa/pip-audit/issues/472
+                    and all([any(r.specifier.filter((candidate.version,))) for r in requirements])
+                    # HACK(ww): Additionally check that each candidate's name matches the
+                    # expected project name (identifier).
+                    # This technically shouldn't be required, but parsing distribution names
+                    # from package indices is imprecise/unreliable when distribution filenames
+                    # are PEP 440 compliant but not normalized.
+                    # See: https://github.com/pypa/packaging/issues/527
+                    and candidate.name == identifier
+                ],
+                key=attrgetter("version", "is_wheel"),
+                reverse=True,
+            )
+        except PyPINotFoundError as e:
+            yield SkippedCandidate(name=identifier, skip_reason=str(e))
+            return
+
+        logger.debug(f"{identifier} has candidates: {candidates}")
+
         # If we have multiple candidates for a single version and some are wheels,
         # yield only the wheels. This keeps us from wasting a large amount of
         # dependency search time when comparing wheels against source distributions.
@@ -406,12 +500,24 @@ class PyPIProvider(AbstractProvider):
         """
         See `resolvelib.providers.AbstractProvider.is_satisfied_by`.
         """
-        return candidate.version in requirement.specifier
+        if isinstance(candidate, SkippedCandidate):
+            return True
+
+        # See the NOTE in find_matches: we use `filter(...)` because of its
+        # special casing around prereleases.
+        return any(
+            requirement.specifier.filter(
+                (candidate.version,),
+            )
+        )
 
     def get_dependencies(self, candidate: Any) -> Any:
         """
         See `resolvelib.providers.AbstractProvider.get_dependencies`.
         """
+        if isinstance(candidate, SkippedCandidate):
+            return []
+
         return candidate.dependencies
 
 
